#include <sys/types.h>

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "dir.h"
#include "file.h"

void
dir_free(struct dir *dir)
{
	/* next head */
	struct dir_entry *next;

	if (!dir->head) return;
	for (next = dir->head->next; next; next = next->next) {
		if (dir->head->name)
			free(dir->head->name);
		free(dir->head);
		dir->head = next;
	}

	/* dir->head will be 0 */
}

int
dir_open(int fs_fd, uint32_t inode, struct dir *dir)
{
	uint8_t buf[SZ_BLOCK];

	struct file	file;

	off_t			file_pos; /* pos in the file */
	struct dir_entry	*tail;
	ssize_t			bytes;
	size_t			i;
	uint32_t		cur_inode;
	size_t			len; /* len of current name */
	size_t			sz_rec;
	size_t			sz; /* bytes in buf */
	int			status;

	if (file_open(fs_fd, inode, &file) == -1) {
		fprintf(stderr, "dir_open(): failed to open directory file\n");
		return -1;
	}

	dir->head = 0;
	tail = 0;
	i = 0;
	file_pos = 0;
	sz = 0;
	dir->perms = file.perms;
	dir->uid = file.uid;
	dir->gid = file.gid;
	while (file_pos < file.sz) {
		if (i == sz) {
			/* runs the first time through the loop, and whenever
			 * a block has been digested; directory entries
			 * don't span between two blocks */
			i = 0;
			sz = 0;
			do {
				bytes = file_read(&file, buf + sz,
				    sizeof(buf) - sz);
				if (bytes == -1) {
					fprintf(stderr,
					    "dir_open(): failed to read "
					    "from directory\n");
					status = -1;
					goto fail;
				} else if (!bytes) {
					break;
				}
				sz += bytes;
			} while (sz < sizeof(buf));
		}

		sz_rec = *(uint16_t *)(buf + i + EXT4_DOFF_RECSZ);
		len = buf[i + EXT4_DOFF_NAMELEN];
		cur_inode = *(uint32_t *)(buf + i + EXT4_DOFF_INODE);

		/* an inode of zero means the file at the beginning of a
		 * block was deleted, but there was no previous entry
		 * (otherwise, the previous entry's record length would be
		 * increased); also skip '.' and '..' */
		if ((len == 1 && buf[i + EXT4_DOFF_NAME] == '.') ||
		   (len == 2 && !strncmp(
		   (char *)(buf + i + EXT4_DOFF_NAME), "..", 2)) ||
		   cur_inode == 0) {
			i += sz_rec;
			file_pos += sz_rec;
			continue;
		}

		/* alloc new tail */
		if (tail) {
			tail->next = malloc(sizeof(struct dir_entry));
			if (!tail->next) {
				perror("dir_open(): malloc()");
				status = -1;
				goto fail;
			}
			tail = tail->next;
		} else {
			tail = malloc(sizeof(struct dir_entry));
			if (!tail) {
				perror("dir_open(): malloc()");
				status = -1;
				goto fail;
			}
			dir->head = tail;
		}
		tail->next = 0;

		if (!(tail->name = malloc(len + 1))) {
			perror("dir_open(): malloc()");
			status = -1;
			goto fail;
		}
		memcpy(tail->name, buf + i + EXT4_DOFF_NAME, len);
		tail->name[len] = '\0';
		tail->inode = cur_inode;
		tail->type = (enum file_type)buf[i + EXT4_DOFF_TYPE];

		i += sz_rec;
		file_pos += sz_rec;
	}

	dir->cur = 0;

	status = 0;
fail:
	file_free(&file);
	if (status == -1) dir_free(dir);
	return status;
}

struct dir_entry *
dir_read(struct dir *dir)
{
	if (!dir->cur) dir->cur = dir->head;
	else dir->cur = dir->cur->next;
	return dir->cur;
}

void
dir_rewind(struct dir *dir)
{
	dir->cur = 0;
}
