#include <sys/types.h>
#include <arpa/inet.h>

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "ext4.h"
#include "file.h"

enum advance {
	ADV_SUCCESS,	/* advanced */
	ADV_DONE,	/* just completed */
	ADV_FAIL	/* previously completed */
};

enum block_entry {
	TYPE_UNDEFINED = 0,
	TYPE_LEAF,
	TYPE_BRANCH
};

struct extent_leaf {
	uint64_t start;
	uint16_t count;
	uint16_t pos;
};

struct blocks_extent {
	union {
		struct extent_leaf *leaves;
		struct blocks_extent *branches;
	} _union;

#define		ext_leaves	_union.leaves /* TYPE_LEAF */
#define		ext_branches	_union.branches /* TYPE_BRANCH */

	uint16_t		count;
	uint16_t		pos;
	enum block_entry	type;
};

struct blocks_plain {
	struct blocks_plain	*next;
	uint32_t		*buf;
	uint16_t	    	count;
	uint16_t		pos;
};

static int	extent_init(int, const uint8_t *, struct blocks_extent *);
static void	extent_free(struct blocks_extent *);
static enum advance
		extent_advance(struct blocks_extent *, uint64_t *);

static int	plain_init(int, const uint8_t *, struct blocks_plain *);
static struct blocks_plain *
		plain_addbranch(int, uint32_t, struct blocks_plain *,
		    int);
static void	plain_free(struct blocks_plain *);

static int
extent_init(int fd, const uint8_t *buf, struct blocks_extent *ext)
{
	uint8_t			readbuf[SZ_BLOCK];
	uint64_t		block;
	size_t			i;
	uint16_t		depth;

	if (*(uint16_t *)(buf + EXT4_EHOFF_MAGIC) != EXT4_MAGIC) {
		fprintf(stderr, "extent_init(): no extent magic\n");
		return -1;
	}

	ext->count = *(uint16_t *)(buf + EXT4_EHOFF_COUNT);
	ext->pos = 0;
	depth = *(uint16_t *)(buf + EXT4_EHOFF_DEPTH);
	buf += EXT4_SZ_EXTHDR;

	if (depth) {
		ext->type = TYPE_BRANCH;
		ext->ext_branches = malloc(
		    sizeof(struct blocks_extent) * ext->count);
		if (!ext->ext_branches) {
			perror("extent_init(): malloc()");
			extent_free(ext);
			return -1;
		}

		for (i = 0; i < ext->count; i++) {
			block = (off_t)*(uint16_t *)(buf + EXT4_EIOFF_LEAFHI)
			    << 32 | *(uint32_t *)(buf + EXT4_EIOFF_LEAFLO);
			if (read_block(fd, block, readbuf) == -1) {
				extent_free(ext);
				return -1;
			}

			if (extent_init(fd, readbuf, ext->ext_branches + i)
			    == -1) {
				extent_free(ext);
				return -1;
			}

			buf += EXT4_SZ_EXTIDX;
		}
	} else {
		ext->type = TYPE_LEAF;
		ext->ext_leaves = malloc(
		    sizeof(struct extent_leaf) * ext->count);
		if (!ext->ext_leaves) {
			perror("extent_init(): malloc()");
			extent_free(ext);
			return -1;
		}

		for (i = 0; i < ext->count; i++) {
			ext->ext_leaves[i].start =
			    (off_t)*(uint16_t *)(buf + EXT4_EOFF_BLKHI) << 32 |
			    *(uint32_t *)(buf + EXT4_EOFF_BLKLO);
			ext->ext_leaves[i].count =
			    *(uint16_t *)(buf + EXT4_EOFF_COUNT);
			ext->ext_leaves[i].pos = 0;

			buf += EXT4_SZ_EXTENT;
		}
	}
	return 0;
}

static void
extent_free(struct blocks_extent *ext)
{
	uint16_t i;
	if (ext->type == TYPE_LEAF) {
		free(ext->ext_leaves);
	} else if (ext->type == TYPE_BRANCH) {
		for (i = 0; i < ext->count; i++)
			extent_free(ext->ext_branches + i);
		free(ext->ext_branches);
	}
	ext->type = TYPE_UNDEFINED;
}

static enum advance
extent_advance(struct blocks_extent *ext, uint64_t *out_block)
{
	if (ext->pos == ext->count)
		return ADV_FAIL;

	if (ext->type == TYPE_LEAF) {
		*out_block = ext->ext_leaves[ext->pos].start +
		    ext->ext_leaves[ext->pos].pos;

		if (++ext->ext_leaves[ext->pos].pos ==
		    ext->ext_leaves[ext->pos].count) {
			/* move to next extent */
			if (++ext->pos == ext->count)
				return ADV_DONE;
		}
		return ADV_SUCCESS;
	}

	switch (extent_advance(ext->ext_branches + ext->pos, out_block)) {
	case ADV_SUCCESS: return ADV_SUCCESS;
	case ADV_FAIL: return ADV_FAIL;
	case ADV_DONE:
		if (++ext->pos == ext->count)
			return ADV_DONE;
		return ADV_SUCCESS;
	default:
		return ADV_FAIL;
	}
}

/* indirection is the count of levels before the leaves start; if it is 0,
 * then 'block' contains data block numbers; returns the new tail */
static struct blocks_plain *
plain_addbranch(int fd, uint32_t block, struct blocks_plain *tail,
    int indirection)
{
	uint32_t		blocks[SZ_BLOCK];
	struct blocks_plain	*newtail;
	struct blocks_plain	*tmp;
	uint16_t		count;
	uint16_t		i;

	if (read_block(fd, block, (uint8_t *)blocks) == -1) {
		fprintf(stderr, "plain_addbranch(): read_block() failed\n");
		return 0;
	}

	for (count = 0; count < ADDRS_PER_BLOCK && blocks[count]; count++);

	if (!indirection) {
		newtail = malloc(sizeof(struct blocks_plain));
		if (!newtail) {
			perror("plain_addbranch(): malloc()");
			return 0;
		}

		newtail->count = count;

		newtail->buf = malloc(4 * newtail->count);
		if (!newtail->buf) {
			perror("plain_addbranch(): malloc()");
			free(newtail);
			return 0;
		}

		newtail->next = 0;
		memcpy(newtail->buf, blocks, 4 * newtail->count);
		newtail->pos = 0;
		tail->next = newtail;
		return newtail;
	}

	newtail = tail;
	for (i = 0; i < count; i++) {
		tmp = plain_addbranch(fd, blocks[i], newtail, indirection-1);
		if (!tmp) {
			if (tail->next) {
				/* make everything as it was */
				plain_free(tail->next);
				free(tail->next);
				tail->next = 0;
			}
			return 0;
		}
		newtail = tmp;

		if (newtail->count != ADDRS_PER_BLOCK) break;
	}

	return newtail;
}

static int
plain_init(int fd, const uint8_t *buf, struct blocks_plain *pln)
{
	uint32_t		*blocks = (uint32_t *)buf;
	struct blocks_plain	*newtail;
	struct blocks_plain	*tail;

	pln->next = 0;
	/* find first non-zero NDIR block */
	for (pln->count = 0;
	    pln->count < EXT4_NDIR_BLOCKS && blocks[pln->count];
	    pln->count++);
	pln->pos = 0;

	pln->buf = malloc(4 * pln->count);
	if (!pln->buf) {
		perror("plain_init(): malloc()");
		return -1;
	}
	memcpy(pln->buf, buf, pln->count * 4);

	/* no IND */
	if (pln->count != EXT4_NDIR_BLOCKS || !blocks[EXT4_IND_BLOCK])
		return 0;

	/* add IND */
	newtail = plain_addbranch(fd, blocks[EXT4_IND_BLOCK], pln, 0);
	if (!newtail) {
		plain_free(pln);
		return -1;
	}
	tail = newtail;

	/* no DIND */
	if (tail->count != ADDRS_PER_BLOCK && !blocks[EXT4_DIND_BLOCK])
		return 0;

	/* add DIND */
	newtail = plain_addbranch(fd, blocks[EXT4_DIND_BLOCK], tail, 1);
	if (!newtail) {
		plain_free(pln);
		return -1;
	}
	tail = newtail;

	/* no TIND */
	if (tail->count != ADDRS_PER_BLOCK && !blocks[EXT4_TIND_BLOCK])
		return 0;

	/* add TIND */
	newtail = plain_addbranch(fd, blocks[EXT4_DIND_BLOCK], tail, 1);
	if (!newtail) {
		plain_free(pln);
		return -1;
	}
	return 0;
}

static void
plain_free(struct blocks_plain *pln)
{
	struct blocks_plain *head;
	struct blocks_plain *nexthead;

	if (pln->buf) {
		free(pln->buf);
		pln->buf = 0;
	}

	head = pln->next;
	while (head) {
		nexthead = head->next;
		if (head->buf) free(head->buf);
		free(head);
		head = nexthead;
	}

	pln->next = 0;
	pln->count = 0;
	pln->pos = 0;
}

/* will modify 'file' even if error is encountered */
int
file_open(int fs_fd, uint32_t inode, struct file *file)
{
	uint8_t		buf[SZ_INODE];

	uint8_t		*blocks;
	ssize_t		bytes;
	ssize_t		bytes_left;
	uint32_t	flags;
	uint16_t	mode;

	file->off_inode = get_inode_offset(inode);

	if (file->off_inode == -1) {
		fprintf(stderr, "file_open(): inode 0x%lx outside bounds\n",
		    (long unsigned)inode);
		return -1;
	}

	if (lseek(fs_fd, file->off_inode, SEEK_SET) == -1) {
		perror("file_open(): lseek()");
		return -1;
	}

	bytes_left = SZ_INODE;
	do {
		bytes = read(fs_fd, buf + (SZ_INODE - bytes_left), bytes_left);
		if (bytes == -1 && errno == EINTR) continue;
		if (bytes == -1) {
			perror("file_open(): read()");
			return -1;
		}
		bytes_left -= bytes;
	} while (bytes_left);

	file->fd = fs_fd;

	/* block details */
	blocks = buf + EXT4_IOFF_BLOCKS;
	file->block_count = *(uint32_t *)(buf + EXT4_IOFF_NUMBLOCKS);
	file->block_pos = 0;

	/* inode details */
	file->sz = (off_t)*(uint32_t *)(buf + EXT4_IOFF_SZHI) << 32 |
	    *(uint32_t *)(buf + EXT4_IOFF_SZLO);
	file->inode = inode;
	file->uid = *(uint16_t *)(buf + EXT4_IOFF_UID);
	file->gid = *(uint16_t *)(buf + EXT4_IOFF_GID);
	flags = *(uint32_t *)(buf + EXT4_IOFF_FLAGS);
	mode = *(uint16_t *)(buf + EXT4_IOFF_MODE);
	file->perms = mode & 07777;
	file->type = mode2filetype(mode);

	/* buffer details */
	file->buf_pos = 0;
	file->buf_sz = 0;
	file->eof = 0;

	if (!file->block_count) {
		/* do a 'read', since it's free */
		file->block_type = BLK_NONE;
		memcpy(file->buf, blocks, file->sz);
		file->buf_sz = file->sz;
		file->eof = 1;

	} else if (flags & EXT4_EXTENTS_FL &&
	    *(uint16_t *)(blocks + EXT4_EHOFF_MAGIC) == EXT4_MAGIC) {

		/* extent flag set AND magic present => use extents */
		file->pos_extent = malloc(sizeof(struct blocks_extent));
		if (!file->pos_extent) {
			perror("file_open(): malloc()");
			return -1;
		}

		file->block_type = BLK_EXTENT;
		if (extent_init(fs_fd, blocks, file->pos_extent) == -1) {
			fprintf(stderr,
			    "file_open(): extent_init() failed\n");
			free(file->pos_extent);
			file->block_type = BLK_UNDEFINED;
			return -1;
		}
	} else {
		file->pos_plain = malloc(sizeof(struct blocks_plain));
		if (!file->pos_plain) {
			perror("file_open(): malloc()");
			return -1;
		}

		file->block_type = BLK_PLAIN;
		if (plain_init(fs_fd, blocks, file->pos_plain) == -1) {
			fprintf(stderr,
			    "file_open(): plain_init() failed\n");
			free(file->pos_plain);
			file->block_type = BLK_UNDEFINED;
			return -1;
		}
	}

	return 0;
}

void
file_free(struct file *file)
{
	if (file->block_type == BLK_EXTENT) {
		extent_free(file->pos_extent);
		free(file->pos_extent);
	} else if (file->block_type == BLK_PLAIN) {
		if (file->pos_plain) {
			plain_free(file->pos_plain);
			free(file->pos_plain);
		}
	}
	file->block_type = BLK_UNDEFINED;
}

ssize_t
file_read(struct file *file, uint8_t *buf, size_t sz)
{
	/* block address to read */
	uint64_t		block;
	off_t			past_bytes;
	ssize_t			bytes;
	struct blocks_plain	*newhead;
	int			retval;

	if (file->buf_pos < file->buf_sz) {
		/* some bytes left in the buffer */
		bytes = file->buf_sz - file->buf_pos;
		if (sz < bytes) bytes = sz;
		memcpy(buf, file->buf + file->buf_pos, bytes);
		file->buf_pos += bytes;
		return bytes;
	}

	if (file->eof) {
		/* no more blocks available */
		return 0;
	}

	/* determine next block, advance */
	if (file->block_type == BLK_PLAIN) {
		block = file->pos_plain->buf[file->pos_plain->pos++];
		if (file->pos_plain->pos == file->pos_plain->count) {
			newhead = file->pos_plain->next;

			/* remove just first */
			file->pos_plain->next = 0;
			plain_free(file->pos_plain);
			free(file->pos_plain);

			file->pos_plain = newhead;

			if (!file->pos_plain)
				file->eof = 1;
		}
	} else if (file->block_type == BLK_EXTENT) {
		block = 0;
		switch (extent_advance(file->pos_extent, &block)) {
		case ADV_FAIL:
			assert(0);
			break;
		case ADV_DONE:
			file->eof = 1;
		case ADV_SUCCESS:;
		}
	}

	retval = read_block(file->fd, block, file->buf);
	if (retval == -1) {
		fprintf(stderr, "file_read(): read_block()\n");
		return -1;
	}

	/* how many bytes in this block? */
	past_bytes = (off_t)file->block_pos * SZ_BLOCK;
	if (file->sz - past_bytes < SZ_BLOCK) {
		/* partial block */
		file->buf_sz = file->sz - past_bytes;
		/* I don't know why this doesn't always get set to 1 at EOF
		 * above */
		file->eof = 1;
	}
	else
		file->buf_sz = SZ_BLOCK;
	file->buf_pos = 0;

	file->block_pos++;

	/* copy to caller */
	bytes = file->buf_sz;
	if (sz < bytes) bytes = sz;
	memcpy(buf, file->buf, bytes);
	file->buf_pos += bytes;

	return bytes;
}

const char *
filetype2str(enum file_type t)
{
	switch (t) {
	case FT_UNKNOWN: return "unknown";
	case FT_REG: return "regular file";
	case FT_DIR: return "directory";
	case FT_CHRDEV: return "character device";
	case FT_BLKDEV: return "block device";
	case FT_FIFO: return "fifo";
	case FT_SOCK: return "socket";
	case FT_SYMLINK: return "symbolic link";
	default:
		/* there is no default */
		fprintf(stderr, "?? type %u\n", t);
		return 0;
	}
}

enum file_type
mode2filetype(uint16_t mode)
{
	switch (mode >> 12) {
	case 0xc: return FT_SOCK;
	case 0xa: return FT_SYMLINK;
	case 0x8: return FT_REG;
	case 0x6: return FT_BLKDEV;
	case 0x4: return FT_DIR;
	case 0x2: return FT_CHRDEV;
	case 0x1: return FT_FIFO;
	default: return FT_UNKNOWN;
	}
}

/* returns -1 (seek or read error) or SZ_BLOCK (success) */
int
read_block(int fs_fd, uint64_t block, uint8_t *buf)
{
	ssize_t		bytes;
	ssize_t		tot;

	if (lseek(fs_fd, (off_t)block * SZ_BLOCK, SEEK_SET) == -1) {
		perror("read_block(): lseek()");
		return -1;
	}

	tot = 0;
	do {
		bytes = read(fs_fd, buf + tot, SZ_BLOCK - tot);
		if (bytes == -1) {
			if (errno == EINTR) continue;
			perror("read_block(): read()");
			return -1;
		}
		tot += bytes;
	} while (tot < SZ_BLOCK);
	return SZ_BLOCK;
}
