#include <arpa/inet.h>

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "log.h"

struct log_entry {
	char *path;
	uint16_t fields;

	/* LT_LINK */
	char *link;

	/* LT_PERM */
	uint16_t perm;

	/* LT_UID */
	uint16_t uid;

	/* LT_GID */
	uint16_t gid;
};

const uint16_t LT_LINK = 0x1;
const uint16_t LT_PERM = 0x2;
const uint16_t LT_UID = 0x4;
const uint16_t LT_GID = 0x8;

int
log_write(int fd, struct log_entry *entry)
{
	uint8_t		*buf;
	ssize_t		bytes;
	size_t		dstlen = 0;
	size_t		pathlen;
	uint8_t		*pos;
	size_t		written;
	uint16_t	rec_len;
	uint16_t	t16;

	rec_len = 6;
	pathlen = strlen(entry->path);
	rec_len += pathlen;
	if (entry->fields & LT_LINK) {
		dstlen = strlen(entry->link);
		rec_len += dstlen + 2;
	}
	if (entry->fields & LT_PERM)
		rec_len += 2;
	if (entry->fields & LT_UID)
		rec_len += 2;
	if (entry->fields & LT_GID)
		rec_len += 2;

	if (!(buf = malloc(rec_len))) {
		perror("log_write(): malloc()");
		return -1;
	}

	*(uint16_t *)(buf + 0) = htons(rec_len);
	*(uint16_t *)(buf + 2) = htons(entry->fields);
	*(uint16_t *)(buf + 4) = htons(strlen(entry->path));
	pos = buf + 6;
	memcpy(pos, entry->path, pathlen);
	pos += pathlen;

	if (entry->fields & LT_LINK) {
		t16 = htons(dstlen);
		memcpy(pos, (uint8_t *)&t16, 2);
		pos += 2;
		memcpy(pos, entry->link, dstlen);
		pos += dstlen;
	}
	if (entry->fields & LT_PERM) {
		t16 = htons(entry->perm);
		memcpy(pos, (uint8_t *)&t16, 2);
		pos += 2;
	}
	if (entry->fields & LT_UID) {
		t16 = htons(entry->uid);
		memcpy(pos, (uint8_t *)&t16, 2);
		pos += 2;
	}
	if (entry->fields & LT_GID) {
		t16 = htons(entry->gid);
		memcpy(pos, (uint8_t *)&t16, 2);
		pos += 2;
	}

	written = 0;
	do {
		bytes = write(fd, buf + written, rec_len - written);
		if (bytes == -1) {
			if (errno == EINTR) continue;
			perror("log_write(): write()");
			free(buf);
			return -1;
		}
		written += bytes;
	} while (written < rec_len);

	free(buf);
	return 0;
}

struct log_entry *
log_read(int fd)
{
	uint8_t		*buf = 0;
	struct log_entry *entry = 0;
	uint8_t		*pos;
	ssize_t		bytes;
	size_t		sz;
	uint16_t	rec_len;
	uint16_t	t16; /* temp-16 */
	uint8_t		error;

	if (!(entry = malloc(sizeof(struct log_entry)))) {
		perror("log_read(): malloc()");
		return 0;
	}
	entry->path = 0;
	entry->link = 0;

	sz = 0;
	do {
		bytes = read(fd, (uint8_t *)&rec_len + sz, 2 - sz);
		if (bytes == -1) {
			if (errno == EINTR) continue;
			perror("log_read(): read()");
			error = 1;
			goto fail;
		} else if (bytes == 0) {
			if (!sz) return 0;
			fprintf(stderr, "log_read(): premature EOF\n");
			error = 1;
			goto fail;
		}
		sz += bytes;
	} while (sz < 2);
	rec_len = ntohs(rec_len);

	buf = malloc(rec_len - 2);
	if (!buf) {
		perror("log_read(): malloc()");
		error = 1;
		goto fail;
	}

	sz = 0;
	do {
		bytes = read(fd, buf + sz, rec_len - 2 - sz);
		if (bytes == -1) {
			if (errno == EINTR) continue;
			perror("log_read(): read()");
			error = 1;
			goto fail;
		} else if (!bytes) {
			fprintf(stderr, "log_read(): premature EOF\n");
			error = 1;
			goto fail;
		}
		sz += bytes;
	} while (sz < rec_len - 2);

	entry->fields = ntohs(*(uint16_t *)buf);
	t16 = ntohs(*(uint16_t *)(buf + 2));

	if (!(entry->path = malloc(t16 + 1))) {
		perror("log_read(): malloc()");
		error = 1;
		goto fail;
	}

	pos = buf + 4;
	memcpy(entry->path, pos, t16);
	entry->path[t16] = '\0';
	pos += t16;

	if (entry->fields & LT_LINK) {
		memcpy((uint8_t *)&t16, pos, 2);
		t16 = ntohs(t16);
		pos += 2;
		if (!(entry->link = malloc(t16 + 1))) {
			perror("log_read(): malloc()");
			error = 1;
			goto fail;
		}
		memcpy(entry->link, pos, t16);
		entry->link[t16] = '\0';
		pos += t16;
	}
	if (entry->fields & LT_PERM) {
		memcpy((uint8_t *)&t16, pos, 2);
		entry->perm = ntohs(t16);
		pos += 2;
	}
	if (entry->fields & LT_UID) {
		memcpy((uint8_t *)&t16, pos, 2);
		entry->uid = ntohs(t16);
		pos += 2;
	}
	if (entry->fields & LT_GID) {
		memcpy((uint8_t *)&t16, pos, 2);
		entry->gid = ntohs(t16);
		pos += 2;
	}

	error = 0;
fail:
	if (buf) free(buf);
	if (error) {
		if (entry) {
			if (entry->path) free(entry->path);
			if (entry->link) free(entry->link);
			free(entry);
		}
		return 0;
	}

	return entry;
}

struct log_entry *
log_init(const char *path)
{
	struct log_entry *entry;
	size_t len;

	if (!(entry = malloc(sizeof(struct log_entry)))) {
		perror("log_init(): malloc()");
		return 0;
	}

	len = strlen(path);
	if (!(entry->path = malloc(len + 1))) {
		perror("log_init(): malloc()");
		free(entry);
		return 0;
	}
	memcpy(entry->path, path, len + 1);
	entry->fields = 0;
	return entry;
}

int
log_addlink(struct log_entry *entry, const char *dst)
{
	char *realloced;
	size_t len;

	if (!(entry->fields & LT_LINK))
		entry->link = 0;

	len = strlen(dst);
	if (!(realloced = realloc(entry->link, len + 1))) {
		perror("log_addlink(): realloc()");
		return -1;
	}
	entry->link = realloced;

	entry->fields |= LT_LINK;
	memcpy(entry->link, dst, len + 1);
	return 0;
}

void
log_addperm(struct log_entry *entry, uint16_t perm)
{
	entry->fields |= LT_PERM;
	entry->perm = perm;
}

void
log_adduid(struct log_entry *entry, uint16_t uid)
{
	entry->fields |= LT_UID;
	entry->uid = uid;
}

void
log_addgid(struct log_entry *entry, uint16_t gid)
{
	entry->fields |= LT_GID;
	entry->gid = gid;
}

const char *
log_getpath(const struct log_entry *e)
{
	return e->path;
}

const char *
log_getlink(const struct log_entry *e)
{
	if (e->fields & LT_LINK) return e->link;
	return 0;
}

uint16_t
log_getperm(const struct log_entry *e)
{
	if (e->fields & LT_PERM) return e->perm;
	return LOG_UNDEFINED;
}

uint16_t
log_getuid(struct log_entry *e)
{
	if (e->fields & LT_UID) return e->uid;
	return LOG_UNDEFINED;
}

uint16_t
log_getgid(struct log_entry *e)
{
	if (e->fields & LT_GID) return e->gid;
	return LOG_UNDEFINED;
}

void
log_free(struct log_entry *entry)
{
	if (entry->fields & LT_LINK) free(entry->link);
	free(entry->path);
	free(entry);
}
