#include <sys/types.h>
#include <sys/stat.h>

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "dir.h"
#include "file.h"
#include "log.h"
#include "save.h"

static char	*path_pop(char[PATHLEN]);
static void	path_push(char[PATHLEN], const char *);

static int	save_link(int, uint32_t, const char *, int, uint8_t,
		    char[PATHLEN], off_t *);

static char *
path_pop(char p[PATHLEN])
{
	char *slash = strrchr(p, '/');
	if (!slash) slash = p;
	*slash = '\0';
	return slash + 1;
}

static void
path_push(char p[PATHLEN], const char *name)
{
	size_t plen = strlen(p);
	size_t nlen = strlen(name);
	assert(plen + nlen + 1 < PATHLEN);
	if (*p)
		p[plen++] = '/';
	memcpy(p + plen, name, nlen + 1);
}

/* dirname is the name of the directory with inode 'inode'; fullpath is
 * the path to the file that *contains* dirname, and should never end with
 * a '/'; total size of all operations will be written to out_total_size if
 * non-null */
int
save_dir(int fd, uint32_t inode, const char *dirname, int logfd,
    uint8_t pretend, char fullpath[PATHLEN], off_t *out_total_size)
{
	struct dir	dir;

	off_t			bytes;
	struct dir_entry	*entry;
	struct log_entry	*log = 0;
	int			status;
	uint8_t			opened = 0;

	if (!pretend && chdir(dirname) == -1) {
		perror("save_dir(): chdir()");
		return -1;
	}

	if (dir_open(fd, inode, &dir) == -1) {
		fprintf(stderr, "save_dir(): failed to open dir %s\n",
		    dirname);
		status = -1;
		goto cleanup;
	}
	opened = 1;

	path_push(fullpath, dirname);

	if (out_total_size)
		*out_total_size = 0;
	while ((entry = dir_read(&dir))) {
		switch (entry->type) {
		case FT_DIR:
			if (entry->inode == 0x1389d2)
				/* the dreaded chromium source */
				break;

			if (!pretend && mkdir(entry->name, S_IRWXU |
			    S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) == -1) {
				if (errno != EEXIST) {
					perror("save_dir(): mkdir()");
					break;
				}
			}

			if (logfd != -1) {
				/* log the uid, gid, and permissions */
				path_push(fullpath, entry->name);
				log = log_init(fullpath);
				path_pop(fullpath);
				if (!log) {
					fprintf(stderr,
					    "save_dir(): failed to "
					    "create logging entry\n");
					status = -1;
					goto cleanup;
				}
				log_addperm(log, dir.perms);
				log_adduid(log, dir.uid);
				log_addgid(log, dir.gid);
				if (log_write(logfd, log) == -1) {
					fprintf(stderr,
					    "save_dir(): failed to "
					    "write log\n");
					status = -1;
					goto cleanup;
				}
				log_free(log);
				log = 0;
			}

			bytes = 0;
			if (save_dir(fd, entry->inode, entry->name, logfd,
			    pretend, fullpath, &bytes) == -1) {
				break;
			}

			if (out_total_size)
				*out_total_size += bytes;

			break;
		case FT_REG:
			bytes = 0;
			if (save_file(fd, entry->inode, entry->name, logfd,
			    pretend, fullpath, &bytes) == -1) {
				fprintf(stderr,
				    "failed in dir w/inode=0x%lx\n",
				    (long unsigned)inode);
			}
			if (out_total_size)
				*out_total_size += bytes;
			break;
		case FT_SYMLINK:
			bytes = 0;
			if (save_link(fd, entry->inode, entry->name, logfd,
			    pretend, fullpath, &bytes) == -1) {
				fprintf(stderr,
				    "failed in dir w/inode=0x%lx\n",
				    (long unsigned)inode);
			}
			if (out_total_size)
				*out_total_size += bytes;
			break;
		default:
#ifdef DEBUG
			fprintf(stderr, "inode %lu, entry offset %lu\n",
			    (long unsigned)inode,
			    (long unsigned)entry->offset);
#endif
			fprintf(stderr, "save_dir(): "
			    "could not back up .../%s/%s: "
			    "unsupported type %s\n",
			    dirname, entry->name, filetype2str(entry->type));
			break;
		}
	}

	status = 0;
	path_pop(fullpath);
cleanup:
	if (log) log_free(log);
	if (opened) dir_free(&dir);
	if (!pretend && chdir("..") == -1) {
		perror("save_dir(): chdir(\"..\")");
		status = -1;
	}
	return status;
}

int
save_file(int fsfd, uint32_t inode, const char *name, int logfd,
    uint8_t pretend, char fullpath[PATHLEN], off_t *out_total_size)
{
	struct file	file;
	uint8_t		buf[SZ_BLOCK];

	size_t		bufsize;
	struct log_entry *log = 0;
	size_t		written;
	ssize_t		bytes;
	int		fd = -1;
	int		status;

	if (file_open(fsfd, inode, &file) == -1) {
		fprintf(stderr, "save_file(): failed to open %s\n",
		    name);
		status = -1;
		goto cleanup;
	}

	if (out_total_size)
		*out_total_size = file.sz;

	/* log the uid, gid, and permissions */
	path_push(fullpath, name);
	if (logfd != -1) {
		log = log_init(fullpath);
	}
	printf("FILE %s (inode: 0x%lx size: %lld)\n",
	    fullpath, (long unsigned)inode, file.sz);
	path_pop(fullpath);

	if (logfd != -1) {
		if (!log) {
			fprintf(stderr, "save_file(): failed to create "
			    "logging entry\n");
			status = -1;
			goto cleanup;
		}
		log_addperm(log, file.perms);
		log_adduid(log, file.uid);
		log_addgid(log, file.gid);
		if (log_write(logfd, log) == -1) {
			fprintf(stderr, "save_file(): failed to write log\n");
			status = -1;
			goto cleanup;
		}
	}

	if (pretend) {
		status = 0;
		goto cleanup;
	}

	/* overwrite file, for multiple runs */
	fd = open(name, O_TRUNC | O_CREAT | O_WRONLY,
	    S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (fd == -1) {
		fprintf(stderr, "save_file(): open(%s): %s\n",
		    name, strerror(errno));
		status = -1;
		goto cleanup;
	}

	for (;;) {
		bufsize = file_read(&file, buf, sizeof(buf));
		if (bufsize == -1) {
			fprintf(stderr, "save_file(): file_read() failed\n");
			status = -1;
			goto cleanup;
		}

		if (!bufsize) break;

		written = 0;
		do {
			bytes = write(fd, buf + written, bufsize - written);
			if (bytes == -1) {
				if (errno == EINTR) continue;
				perror("save_file(): write()");
				status = -1;
				goto cleanup;
			}
			written += bytes;
		} while (written < bufsize);
	}

	status = 0;
cleanup:
	if (log) log_free(log);
	file_free(&file);
	if (fd != -1) close(fd);
	return status;
}

static int
save_link(int fsfd, uint32_t inode, const char *name, int logfd,
    uint8_t pretend, char fullpath[PATHLEN], off_t *out_total_size)
{
	struct file	file;

	char			*dst = 0;
	struct log_entry	*log = 0;
	ssize_t			bytes;
	int			status;
	size_t			sz;

	if (file_open(fsfd, inode, &file) == -1) {
		fprintf(stderr, "save_link(): failed to open %s\n",
		    name);
		status = -1;
		goto cleanup;
	}

	if (out_total_size)
		*out_total_size = file.sz;

	dst = malloc(file.sz + 1);
	if (!dst) {
		perror("save_link(): malloc()");
		status = -1;
		goto cleanup;
	}

	/* read the link path */
	sz = 0;
	do {
		bytes = file_read(&file, (uint8_t *)dst + sz, file.sz - sz);
		if (bytes == -1) {
			fprintf(stderr, "save_link(): failed to read %s\n",
			    name);
			status = -1;
			goto cleanup;
		}
		sz += bytes;
	} while (sz < file.sz);
	dst[sz] = '\0';

	/* symlinks probably won't work well outside of unix; just log it */
	path_push(fullpath, name);
	if (logfd != -1) {
		log = log_init(fullpath);
	}
	printf("LINK %s (inode: 0x%lx size: %lld)\n",
	    fullpath, (long unsigned)inode, file.sz);
	path_pop(fullpath);

	if (logfd != -1) {
		if (!log) {
			fprintf(stderr, "save_file(): failed to create logging "
			    "entry\n");
			status = -1;
			goto cleanup;
		}
		if (log_addlink(log, dst) == -1) {
			fprintf(stderr, "save_file(): failed to create logging "
			    "entry\n");
			status = -1;
			goto cleanup;
		}
		log_adduid(log, file.uid);
		log_addgid(log, file.gid);
		if (log_write(logfd, log) == -1) {
			fprintf(stderr, "save_file(): failed to write log\n");
			status = -1;
			goto cleanup;
		}
	}

	status = 0;
cleanup:
	if (dst) free(dst);
	if (log) log_free(log);
	file_free(&file);
	return status;
}
